@@ -3,11 +3,11 @@
 ;; Copyright (C) 2011-2014  Guillaume Papin
 
 ;; Author: Guillaume Papin <guillaume.papin@epitech.eu>
-;; Version: 0.1.2
+;; Version: 0.2.0
 ;; URL: https://github.com/Sarcasm/irony-mode
 ;; Compatibility: GNU Emacs 23.x, GNU Emacs 24.x
 ;; Keywords: c, convenience, tools
-;; Package-Requires: ((cl-lib "0.5"))
+;; Package-Requires: ((cl-lib "0.5") (json "1.2"))
 
 ;; This program is free software; you can redistribute it and/or modify
 ;; it under the terms of the GNU General Public License as published by
@@ -50,8 +50,6 @@
 
 ;;; Code:
 
-(autoload 'irony-cdb-load-compile-options "irony-cdb")
-
 (autoload 'irony-completion--enter "irony-completion")
 (autoload 'irony-completion--exit "irony-completion")
 
@@ -118,7 +116,7 @@ The slash is expected at the end."
 
 ;;;###autoload
 (defcustom irony-additional-clang-options nil
-  "Additionnal command line options to pass down to libclang.
+  "Additional command line options to pass down to libclang.
 
 Please, do NOT use this variable to add header search paths, only
 additional warnings or compiler options.
@@ -130,28 +128,16 @@ database."
   :options '("-Wdocumentation")
   :group 'irony)
 
-;;;###autoload
-(defcustom irony-clang-options-updated-hook nil
-  "Normal hook run when the command line options have been updated.
-
-This hook is run when the variables
-`irony-header-search-directories',
-`irony-clang-working-directory', etc, have been updated."
-  ;; TODO: no etc...
-  :type 'hook
-  ;; TODO: :options '(ff-find-other-file/ff-search-directories)
-  :group 'irony)
-
-(defcustom irony-clang-lang-options-alist
+(defcustom irony-lang-compile-option-alist
   '((c++-mode  . "c++")
     (c-mode    . "c")
     (objc-mode . "objective-c"))
-  "Association list of major-mode/lang-option to pass to clang."
+  "Alist to decide the language option to used based on the `major-mode'."
   :type '(alist :key-type symbol :value-type string)
   :group 'irony)
 
 (defcustom irony-cmake-executable "cmake"
-  "The name or path of the CMake executable."
+  "Name or path of the CMake executable."
   :type 'string
   :group 'irony)
 
@@ -164,8 +150,11 @@ CMakeLists.txt used to build the server."
   :type 'directory
   :group 'irony)
 
-(defcustom irony-server-build-dir (concat irony-user-dir "build/")
-  "Build directory for irony-server."
+(defcustom irony-server-build-dir nil
+  "Build directory for irony-server.
+
+If set to nil the default is to create a build directory in
+`temporary-file-directory'/build-irony-server-`(irony-version)'."
   :type 'directory
   :group 'irony)
 
@@ -177,35 +166,6 @@ The irony-server executable is expected to be in
   :type 'directory
   :group 'irony)
 
-(defcustom irony-check-compile-functions
-  '(irony--process-initial-check-compile)
-  "Special hook run when check-compile results are available.
-
-Takes as an argument the number of fatal errors (usually 0 or 1),
-the number of errors and the number of warnings.
-
-Example function:
-
-  (defun my-irony-check-compile-function (nfatals nerrors nwarnings)
-    (message \"%d error(s) and %d warning(s)\" (+ nfatals nerrors)\
- nwarnings))
-
-  (add-hook 'irony-check-compile-functions\
- 'my-irony-check-compile-function)"
-  :type 'hook
-  :options '(irony--process-initial-check-compile)
-  :group 'irony)
-
-(defcustom irony-source-file-extensions '("c"   "cc"
-                                          "C"   "CC"
-                                          "cpp" "cxx" "c++"
-                                          "m"   "mm")
-  "Known file extensions used for source code in C/C++/Obj-C.
-
-Header files extensions should NOT take part of this list."
-  :type '(choice (repeat string))
-  :group 'irony)
-
 
 ;;
 ;; Public/API variables
@@ -217,12 +177,15 @@ Header files extensions should NOT take part of this list."
 ;; read.
 ;;
 
-(defvar irony-header-search-directories nil
-  "Header search directories list for the current buffer.
+;; TODO: make this variable public when the CDB API stabilizes.
+(defvar-local irony--compile-options nil
+  "Compile options for the current file.
 
-Contains the absolute paths to the directories.")
+The compile options used by the compiler to build the current
+buffer file.")
 
-(defvar-local irony-clang-working-directory nil
+;; TODO: make this variable public when the CDB API stabilizes.
+(defvar-local irony--working-directory nil
   "The working directory to pass to libclang, if any.")
 
 
@@ -234,69 +197,9 @@ Contains the absolute paths to the directories.")
 ;;
 ;; -- https://lists.gnu.org/archive/html/emacs-devel/2013-06/msg01129.html
 
-(defvar-local irony--clang-options nil
-  "Compile options for the current file.
-
-The compile options used by the compiler to build the current
-buffer file.")
-
 (defconst irony--eot "\n;;EOT\n"
   "String sent by the server to signal the end of a response.")
 
-(defconst irony-server-eot "\nEOT\n"
-  "The string to send to the server to finish a transmission.")
-
-(defvar irony--server-install-command-history nil)
-
-(defvar-local irony--initial-compile-check-status nil
-  "Non-nil when an initial compile check as already been requested.
-
-Possible values are:
-- nil
-- 'requested, when the compile check for the current buffer has
-  been requested.
-- 'done, when the compile check has been received and processed")
-
-
-;;
-;; Mode
-;;
-
-(defvar irony-mode-map (make-sparse-keymap)
-  "Keymap used in `irony-mode' buffers.")
-
-;;;###autoload
-(define-minor-mode irony-mode
-  "Minor mode for C, C++ and Objective-C, powered by libclang."
-  nil
-  irony-lighter
-  irony-mode-map
-  :group 'irony
-  (if irony-mode
-      (irony-mode-enter)
-    (irony-mode-exit)))
-
-(defun irony-mode-enter ()
-  ;; warn the user about modes such as php-mode who inherits c-mode
-  (when (not (memq major-mode irony-supported-major-modes))
-    (display-warning 'irony "Major mode is unknown to Irony,\
- see `irony-supported-major-modes'."))
-  ;; warn the user about Windows-specific issues
-  (when (eq system-type 'windows-nt)
-    (cond
-     ((version< emacs-version "24.4")
-      (display-warning 'irony "Emacs >= 24.4 expected on Windows."))
-     ((and (boundp 'w32-pipe-read-delay) (> w32-pipe-read-delay 0))
-      (display-warning 'irony "Performance will be bad because a\
- pipe delay is set for this platform (see variable\
- `w32-pipe-read-delay')."))))
-  (unless irony--clang-options
-    (irony-cdb-load-compile-options))
-  (irony-completion--enter))
-
-(defun irony-mode-exit ()
-  (irony-completion--exit))
-
 
 ;;
 ;; Utility functions & macros
@@ -315,7 +218,13 @@ Possible values are:
      (when it
        (progn ,@body))))
 
-(defmacro irony-without-narrowing (&rest body)
+(defun irony--assoc-all (key list)
+  (delq nil (mapcar #'(lambda (c)
+                        (when (equal (car c) key)
+                          c))
+                    list)))
+
+(defmacro irony--without-narrowing (&rest body)
   "Remove the effect of narrowing for the current buffer.
 
 Note: If `save-excursion' is needed for BODY, it should be used
@@ -325,20 +234,7 @@ before calling this macro."
      (widen)
      (progn ,@body)))
 
-(defun irony-version (&optional show-version)
-  "Return the version number of the file irony.el.
-
-If called interactively display the version in the echo area."
-  (interactive (list t))
-  ;; Shamelessly stolen from `company-mode'.
-  (with-temp-buffer
-    (insert-file-contents (find-library-name "irony"))
-    (let ((v (lm-version)))
-      (when show-version
-        (message "irony version: %s" v))
-      v)))
-
-(defun irony-buffer-size-in-bytes ()
+(defun irony--buffer-size-in-bytes ()
   "Return the buffer size, in bytes."
   (1- (position-bytes (point-max))))
 
@@ -366,11 +262,12 @@ to be displayed to the user."
         relative
       abbreviated)))
 
-(defun irony-split-command-line-1 (quoted-str)
+(defun irony--split-command-line-1 (quoted-str)
   "Remove the escaped quotes and backlash from a QUOTED-STR.
 
-Return a list of the final characters in the reverse order, only
-to be consumed by `irony-split-command-line'."
+Return a list of the final characters in the reverse order.
+
+Only to be consumed by `irony--split-command-line'."
   (let ((len (length quoted-str))
         (i 0)
         ch next-ch
@@ -387,7 +284,10 @@ to be consumed by `irony-split-command-line'."
       (cl-incf i))
     result))
 
-(defun irony-split-command-line (cmd-line)
+;; TODO: rewrite the function correctly to handle things like the following:
+;;
+;; "/usr/bin/clang++ -Irelative -DSOMEDEF=\"With spaces, quotes and \\-es.\" <args...>"
+(defun irony--split-command-line (cmd-line)
   "Split CMD-LINE into a list of arguments.
 
 Takes care of double quotes as well as backslash.
@@ -418,7 +318,7 @@ breaks with escaped quotes in compile_commands.json, such as in:
           (unless endq
             (error "Irony: ill formed command line"))
           (let ((quoted-str (substring cmd-line (1+ i) (1+ endq))))
-            (setq cur-arg (append (irony-split-command-line-1 quoted-str)
+            (setq cur-arg (append (irony--split-command-line-1 quoted-str)
                                   cur-arg)
                   i (+ endq 2)))))
        (t                             ;a valid char
@@ -437,38 +337,70 @@ breaks with escaped quotes in compile_commands.json, such as in:
       (setq args (cons (apply 'string (nreverse cur-arg)) args)))
     (nreverse args)))
 
+
+;;
+;; Mode
+;;
+
+(defvar irony-mode-map (make-sparse-keymap)
+  "Keymap used in `irony-mode' buffers.")
+
+;;;###autoload
+(define-minor-mode irony-mode
+  "Minor mode for C, C++ and Objective-C, powered by libclang."
+  nil
+  irony-lighter
+  irony-mode-map
+  :group 'irony
+  (if irony-mode
+      (irony--mode-enter)
+    (irony--mode-exit)))
+
+(defun irony--mode-enter ()
+  ;; warn the user about modes such as php-mode who inherits c-mode
+  (when (not (memq major-mode irony-supported-major-modes))
+    (display-warning 'irony "Major mode is unknown to Irony,\
+ see `irony-supported-major-modes'."))
+  ;; warn the user about Windows-specific issues
+  (when (eq system-type 'windows-nt)
+    (cond
+     ((version< emacs-version "24.4")
+      (display-warning 'irony "Emacs >= 24.4 expected on Windows."))
+     ((and (boundp 'w32-pipe-read-delay) (> w32-pipe-read-delay 0))
+      (display-warning 'irony "Performance will be bad because a\
+ pipe delay is set for this platform (see variable\
+ `w32-pipe-read-delay')."))))
+  (irony-completion--enter))
+
+(defun irony--mode-exit ()
+  (irony-completion--exit))
+
+;;;###autoload
+(defun irony-version (&optional show-version)
+  "Return the version number of the file irony.el.
+
+If called interactively display the version in the echo area."
+  (interactive (list t))
+  ;; Shamelessly stolen from `company-mode'.
+  (with-temp-buffer
+    (insert-file-contents (find-library-name "irony"))
+    (let ((v (lm-version)))
+      (when show-version
+        (message "irony version: %s" v))
+      v)))
+
 
 ;;
 ;; Compile options handling
 ;;
 
-(defun irony-update-command-line-options (cmd-line-options
-                                          &optional
-                                          working-directory)
-  "Setup the command line options to pass down to libclang."
-  (setq irony--clang-options cmd-line-options
-        irony-clang-working-directory working-directory)
-  (run-hooks 'irony-clang-options-updated-hook))
-
-(defun irony--libclang-lang-compile-options ()
-  (irony--awhen (cdr-safe (assq major-mode irony-clang-lang-options-alist))
+(defun irony--lang-compile-option ()
+  (irony--awhen (cdr-safe (assq major-mode irony-lang-compile-option-alist))
     (list "-x" it)))
 
-(defun irony--libclang-compile-options ()
-  "The compile options to send to libclang."
-  ;; TODO: if current buffer has no associated file (will be sent as '-') but is
-  ;; in an existing directory, we will want to add -I (directory-file-name
-  ;; buffer-file-name) to find the relative headers
-  (append
-   (irony--libclang-lang-compile-options)
-   (irony--awhen irony-clang-working-directory
-     (unless (irony--extract-working-directory-option irony--clang-options)
-       (list "-working-directory" it)))
-   irony-additional-clang-options
-   irony--clang-options))
-
 (defun irony--extract-working-directory-option (flags)
-  "Look in FLAGS for the '-working-directory' option, if any."
+  "Return working directory specified on the command line, if
+any."
   (catch 'found
     (while flags
       (let ((flag (car flags)))
@@ -480,11 +412,53 @@ breaks with escaped quotes in compile_commands.json, such as in:
          (t
           (setq flags (cdr flags))))))))
 
+(defun irony--adjust-compile-options ()
+  "The compile options to send to libclang."
+  ;; TODO: if current buffer has no associated file (will be sent as '-') but is
+  ;; in an existing directory, we will want to add -I (directory-file-name
+  ;; buffer-file-name) to find the relative headers
+  (append
+   (irony--lang-compile-option)
+   (irony--awhen irony--working-directory
+     (unless (irony--extract-working-directory-option irony--compile-options)
+       (list "-working-directory" it)))
+   irony-additional-clang-options
+   irony--compile-options))
+
+(defun irony--extract-user-search-paths (compile-options work-dir)
+  "Retrieve the user search paths present in COMPILE-OPTIONS.
+
+Relative paths are expanded to be relative to WORK-DIR.
+
+The returned paths are returned as
+directory (`file-name-as-directory').
+
+Note: WORK-DIR is not used when the compile option
+'-working-directory=<directory>' is detected in COMPILE-OPTIONS."
+  (setq work-dir (or (irony--extract-working-directory-option compile-options)
+                     work-dir))
+  (let (include-dirs opt)
+    (while (setq opt (car compile-options))
+      (cond
+       ((string= "-I" opt)
+        (add-to-list 'include-dirs (nth 1 compile-options) t)
+        (setq compile-options (cddr compile-options)))
+       ((string-prefix-p "-I" opt)
+        (add-to-list 'include-dirs (substring opt 2) t)
+        (setq compile-options (cdr compile-options)))
+       (t
+        (setq compile-options (cdr compile-options)))))
+    (delete-dups (mapcar #'(lambda (path)
+                             (file-name-as-directory
+                              (expand-file-name path work-dir)))
+                         include-dirs))))
+
 
 ;;
 ;; Irony-Server setup
 ;;
 
+(defvar irony--server-install-command-history nil)
 (defun irony--install-server-read-command (command)
   (read-shell-command
    "Install command: " command
@@ -508,16 +482,21 @@ The installation requires CMake and the libclang developpement package."
                  (shell-quote-argument irony-server-source-dir)
                  (shell-quote-argument irony-cmake-executable))))
            (irony--install-server-read-command command))))
-  ;; we need to kill the process to be able to install a new one, at least on
-  ;; Windows
-  (make-directory irony-server-build-dir t)
-  (let ((default-directory irony-server-build-dir))
-    (irony--server-kill-process)
-    (with-current-buffer (compilation-start command nil
-                                            #'(lambda (maj-mode)
-                                                "*irony-server build*"))
-      (setq-local compilation-finish-functions
-                  '(irony--server-install-finish-function)))))
+  (let ((build-dir (or irony-server-build-dir
+                       (concat 
+                        (file-name-as-directory temporary-file-directory)
+                        (file-name-as-directory (format "build-irony-server-%s"
+                                                        (irony-version)))))))
+    (make-directory build-dir t)
+    (let ((default-directory build-dir))
+      ;; we need to kill the process to be able to install a new one,
+      ;; at least on Windows
+      (irony-server-kill)
+      (with-current-buffer (compilation-start command nil
+                                              #'(lambda (maj-mode)
+                                                  "*irony-server build*"))
+        (setq-local compilation-finish-functions
+                    '(irony--server-install-finish-function))))))
 
 (defun irony--server-install-finish-function (buffer msg)
   (if (string= "finished\n" msg)
@@ -573,20 +552,24 @@ list (and undo information is not kept).")
           (process-adaptive-read-buffering nil)
           process)
       (setq process
-            (start-process "Irony" irony--server-buffer
-                           irony--server-executable
-                           "-i"
-                           "--log-file"
-                           (expand-file-name
-                            (format-time-string "irony.%Y-%m-%d_%Hh-%Mm-%Ss.log")
-                            temporary-file-directory)))
+            (start-process-shell-command
+             "Irony"                    ;process name
+             irony--server-buffer       ;buffer
+             (format "%s -i 2> %s"      ;command
+                     (shell-quote-argument irony--server-executable)
+                     (expand-file-name
+                      (format-time-string "irony.%Y-%m-%d_%Hh-%Mm-%Ss.log")
+                      temporary-file-directory))))
       (buffer-disable-undo irony--server-buffer)
       (set-process-query-on-exit-flag process nil)
       (set-process-sentinel process 'irony--server-process-sentinel)
       (set-process-filter process 'irony--server-process-filter)
       process)))
 
-(defun irony--server-kill-process ()
+;;;###autoload
+(defun irony-server-kill ()
+  "Kill the running irony-server process, if any."
+  (interactive)
   (when (and irony--server-process (process-live-p irony--server-process))
     (kill-process irony--server-process)
     (setq irony--server-process nil)))
@@ -605,7 +588,8 @@ list (and undo information is not kept).")
 (defun irony--process-server-response (process response)
   (let ((sexp (read response))
         (callback (irony--server-process-pop-callback process)))
-    (apply (car callback) sexp (cdr callback))))
+    (with-demoted-errors "Warning: %S"
+      (apply (car callback) sexp (cdr callback)))))
 
 (defun irony--server-process-filter (process output)
   "Handle output that come from an irony-server process."
@@ -624,8 +608,8 @@ list (and undo information is not kept).")
             (let ((response (buffer-substring-no-properties (point-min)
                                                             (point))))
               (delete-region (point-min) (point))
-              (setq responses (cons response responses))
-              (goto-char (process-mark process)))))))
+              (setq responses (cons response responses))))
+          (goto-char (process-mark process)))))
     ;; Handle all responses.
     (mapc #'(lambda (r)
               (irony--process-server-response process r))
@@ -642,9 +626,6 @@ list (and undo information is not kept).")
     (process-put p 'irony-callback-stack (cdr callbacks))
     (car callbacks)))
 
-(defun irony--server-process-callback-count (p)
-  (length (process-get p 'irony-callback-stack)))
-
 
 ;;
 ;; Server commands
@@ -659,7 +640,37 @@ If no such file exists on the filesystem the special file '-' is
       buffer-file-name
     "-"))
 
-(defun irony--send-file-request (request callback &rest args)
+(defun irony--send-request (request callback &rest args)
+  (let ((process (irony--get-server-process-create))
+        (argv (cons request args)))
+    (when (and process (process-live-p process))
+      (irony--server-process-push-callback process callback)
+      ;; skip narrowing to compute buffer size and content
+      (irony--without-narrowing
+        (process-send-string process
+                             (format "%s\n"
+                                     (combine-and-quote-strings argv)))))))
+
+(defvar irony--sync-id 0 "ID of next sync request.")
+(defvar irony--sync-result '(-1 . nil)
+  "The car stores the id of the result and the cdr stores the return value.")
+
+(defun irony--sync-request-callback (response id)
+  (setq irony--sync-result (cons id response)))
+
+(defun irony--send-request-sync (request &rest args)
+  "Send a request to irony-server and wait for the result."
+  (let* ((id irony--sync-id)
+         (callback (list #'irony--sync-request-callback id)))
+    (setq irony--sync-id (1+ irony--sync-id))
+    (with-local-quit
+      (let ((process (irony--get-server-process-create)))
+        (apply 'irony--send-request request callback args)
+        (while (not (= id (car irony--sync-result)))
+          (accept-process-output process))
+        (cdr irony--sync-result)))))
+
+(defun irony--send-parse-request (request callback &rest args)
   "Send a request that acts on the current buffer to irony-server.
 
 This concerns mainly irony-server commands that do some work on a
@@ -670,86 +681,140 @@ care of."
                             "--num-unsaved=1"
                             (irony--get-buffer-path-for-server))
                       args))
-        (compile-options (irony--libclang-compile-options)))
+        (compile-options (irony--adjust-compile-options)))
     (when (and process (process-live-p process))
       (irony--server-process-push-callback process callback)
       ;; skip narrowing to compute buffer size and content
-      (irony-without-narrowing
-        (process-send-string process
-                             (format "%s\n%s\n%s\n%d\n"
-                                     (combine-and-quote-strings argv)
-                                     (combine-and-quote-strings compile-options)
-                                     buffer-file-name
-                                     (irony-buffer-size-in-bytes)))
-        (process-send-region process (point-min) (point-max))
-        ;; always make sure to finis with a newline (required by irony-server to
+      (irony--without-narrowing
         ;; always make sure to finish with a newline (required by irony-server
         ;; to play nice with line buffering even when the file doesn't end with
         ;; a newline)
-        (process-send-string process "\n")))))
-
-(defun irony-request-check-compile ()
-  (irony--send-file-request "check-compile"
-                            (list 'irony-check-compile-handler)))
-
-(defun irony-check-compile-handler (errors)
-  (run-hook-with-args 'irony-check-compile-functions
-                      (or (plist-get errors :fatals) 0)
-                      (or (plist-get errors :errors) 0)
-                      (or (plist-get errors :warnings) 0)))
-
-(defun irony--initial-check-compile ()
-  "Check that the current buffer compiles, hinting the user if not.
-
-Ideally this is done only once, when the buffer is first
-opened (or irony-mode first started), just to inform the user if
-he forgot to provide the flags for the current buffer."
-  (unless (or irony--initial-compile-check-status
-              (zerop (buffer-size)))
-    (setq irony--initial-compile-check-status 'requested)
-    (irony-request-check-compile)))
-
-(defun irony--process-initial-check-compile (nfatals nerrors nwarnings)
-  "Display a one-time hint to the user to configure the compile options.
-
-See `irony-check-compile-functions'."
-  (when (and (eq irony--initial-compile-check-status 'requested)
-             (not (zerop (+ nfatals nerrors))))
-    (setq irony--initial-compile-check-status 'done)
-    (let ((help-msg (substitute-command-keys
-                     "Type `\\[irony-cdb-menu]' to configure project"))
-          stats-strings)
-      (unless (zerop nwarnings)
-        (push (concat (number-to-string nwarnings)
-                      " warning"
-                      (unless (eq nwarnings 1) ;plural
-                        "s"))
-              stats-strings))
-      (setq nerrors (+ nfatals nerrors))
-      (unless (zerop nerrors)
-        (push (concat (number-to-string nerrors)
-                      " error"
-                      (unless (eq nerrors 1) ;plural
-                        "s"))
-              stats-strings))
-      (message "[%s] %s" (mapconcat 'identity stats-strings " | ")
-               help-msg))))
+        ;;
+        ;; it is important to send the request atomically rather than using
+        ;; multiple process-send calls. On Windows at least, if the request is
+        ;; not atomic, content from subsequent requests can get intermixed with
+        ;; earlier requests. This may be because of how Emacs behaves when the
+        ;; buffers to communicate with processes are full (see
+        ;; http://www.gnu.org/software/emacs/manual/html_node/elisp/Input-to-Processes.html).
+        (process-send-string process
+                             (format "%s\n%s\n%s\n%d\n%s\n"
+                                     (combine-and-quote-strings argv)
+                                     (combine-and-quote-strings compile-options)
+                                     buffer-file-name
+                                     (irony--buffer-size-in-bytes)
+                                     (buffer-substring (point-min) (point-max))))))))
 
 
 ;;
-;; TODO
+;; Buffer parsing
 ;;
 
-(defun irony-current-directory ()
-  default-directory)
+(defvar-local irony--parse-buffer-state nil
+  "If non-nil, state is of the form (context . status) where:
 
-(defun irony-user-search-paths ()
-  nil)
+- context is `irony--parse-buffer-context'
+- status is one of the following symbol: requested, done")
+
+(defvar-local irony--parse-buffer-callbacks nil)
+(defvar-local irony--parse-buffer-last-results nil
+  "Holds the last parsing results.")
+
+(defun irony--parse-buffer-context ()
+  ;; FIXME: use the ticks of all irony's files that may influence this one?
+  ;; FIXME: compile options should be part of the context? or settings the
+  ;; compile options should flush the context alternatively.
+  (buffer-chars-modified-tick))
+
+(defun irony--buffer-parsed-p ()
+  (equal irony--parse-buffer-state
+         (cons (irony--parse-buffer-context) 'done)))
+
+(defun irony--buffer-parsing-in-progress-p ()
+  (equal irony--parse-buffer-state
+         (cons (irony--parse-buffer-context) 'requested)))
+
+(defun irony--parse-request-handler (result context buffer)
+  (with-current-buffer buffer
+    (let ((callbacks irony--parse-buffer-callbacks)
+          (status (cond
+                   ;; context out-of-date?
+                   ((not (equal context (irony--parse-buffer-context)))
+                    'cancelled)
+                   (result
+                    'success)
+                   (t
+                    'failed))))
+      (setq irony--parse-buffer-last-results (list status)
+            irony--parse-buffer-callbacks nil
+            irony--parse-buffer-state (cons context 'done))
+      (mapc #'(lambda (cb) (funcall cb status)) callbacks))))
+
+;; TODO: provide a synchronous/blocking counterpart, see how
+;; `url-retrieve-synchronously' does it
+(defun irony--parse-buffer-async (callback &optional force)
+  "Parse the current buffer and call CALLACK when done.
+
+Parsing is effectively done only if need, if the buffer hasn't
+changed since the last parsing, CALLBACK is called immediately.
+
+Use FORCE to force a re-parse unconditionally.
+
+Callback is a function that is called with one argument, the
+status of the parsing request, the value is one of the following
+symbol:
+
+- success: parsing the file was a sucess, irony-server has
+  up-to-date information about the buffer
+
+- failed: parsing the file resulted in a failure (file access
+  rights wrong, whatever)
+
+- cancelled: if the request for this callback was superseded by
+  another request or if the callback is out-of-date (but not
+  necessarily superseded by another request)"
+  (when force
+    (setq irony--parse-buffer-state nil))
+  (if (irony--buffer-parsed-p)
+      (apply callback irony--parse-buffer-last-results)
+    (let (obselete-callbacks)
+      (unless (irony--buffer-parsing-in-progress-p)
+        (let ((context (irony--parse-buffer-context)))
+          (setq obselete-callbacks irony--parse-buffer-callbacks
+                irony--parse-buffer-callbacks nil
+                irony--parse-buffer-state (cons context 'requested))
+          (irony--send-parse-request "parse"
+                                    (list 'irony--parse-request-handler context
+                                          (current-buffer)))))
+      (push callback irony--parse-buffer-callbacks)
+      ;; it's safer to call this last, since the function may be called recursively
+      (mapc #'(lambda (cb) (funcall cb 'cancelled)) obselete-callbacks))))
+
+(defun irony-get-type--request-handler (types)
+  (when types
+    (if (cdr types)
+        (if (string= (car types) (cadr types))
+            (message "%s" (car types))
+          (message "%s (aka '%s')" (car types) (cadr types)))
+      (message "%s" (car types)))))
+
+;;;###autoload
+(defun irony-get-type ()
+    "Get the type of symbol under cursor."
+  (interactive)
+  (let ((line (line-number-at-pos))
+        (column (1+ (- (position-bytes (point))
+                       (position-bytes (point-at-bol))))))
+    (irony--parse-buffer-async
+     (lambda (parse-status)
+       (when (eq parse-status 'success)
+         (irony--send-request
+          "get-type"
+          (list 'irony-get-type--request-handler)
+          (number-to-string line)
+          (number-to-string column)))))))
 
 (provide 'irony)
 
-(require 'irony-cdb)
-
 ;; Local Variables:
 ;; byte-compile-warnings: (not cl-functions)
 ;; End:
