@@ -11,8 +11,7 @@
 #ifndef IRONY_MODE_SERVER_COMMAND_H_
 #define IRONY_MODE_SERVER_COMMAND_H_
 
-#include <clang-c/Index.h>
-
+#include "support/CIndex.h"
 #include "support/TemporaryFile.h"
 
 #include <iosfwd>
@@ -30,6 +29,7 @@ struct Command {
     action = Unknown;
     flags.clear();
     file.clear();
+    dir.clear();
     line = 0;
     column = 0;
     unsavedFiles.clear();
@@ -44,6 +44,7 @@ struct Command {
 
   std::vector<std::string> flags;
   std::string file;
+  std::string dir;
   unsigned line;
   unsigned column;
   // pair of (filename, content)
