@@ -17,10 +17,10 @@ namespace support {
 namespace detail {
 
 struct quoted_string_proxy {
-  quoted_string_proxy(std::string &s) : s(s) {
+  quoted_string_proxy(const std::string &s) : s(s) {
   }
 
-  const std::string &s;
+  std::string s;
 };
 
 std::ostream &operator<<(std::ostream &os, const quoted_string_proxy &q) {
@@ -43,7 +43,7 @@ std::ostream &operator<<(std::ostream &os, const quoted_string_proxy &q) {
 
 } // namespace detail
 
-detail::quoted_string_proxy quoted(std::string &s) {
+detail::quoted_string_proxy quoted(const std::string &s) {
   return detail::quoted_string_proxy(s);
 }
 
